#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <openssl/ssl.h>
#include <openssl/safestack.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/x509_vfy.h>

#include "sslauth.h"

//extern X509_STORE *x509st_CRL;
X509_STORE *ssl_X509_STORE_create(char *cpFile, char *cpPath);
int verify_cb_CRL(X509_STORE *x509st_CRL, X509_STORE_CTX *ctx);

void _pam_log(int err, const char *format, ...)
{
    va_list args;

    va_start(args, format);
    openlog("X509_AUTH", LOG_CONS|LOG_PID, LOG_AUTH);
    vsyslog(err, format, args);
    va_end(args);
    closelog();
}

/*
static void _pam_log(int err, const char *format, ...)
{
    va_list args;

    va_start(args, format);
    vprintf(format, args);
    va_end(args);
}

char *crl_dir="/opt/ssl/crl/";
char *crl_file=NULL;
char *ca_dir=NULL;
char *ca_file="/opt/ssl/ca/chain.crt";
char *key_dir="/root";
*/

static int SSLinit(X509_STORE **ca, X509_STORE **crl, const char *cadir, const char *cafile, const char *crldir, const char *crlfile) {
    X509_STORE *nca;
    X509_STORE *ncrl;
    
	// Seems reintialising library don't cause problems
    if (!SSL_library_init())
    {
	_pam_log(LOG_ERR,"Can't initialize OpenSSL library!");
	return -1;
    }

    nca=ssl_X509_STORE_create((char*)cafile,(char*)cadir);
    if (!nca) {
	_pam_log(LOG_ERR,"Can't load CA certificate chain!");
	return -1;
    }

    ncrl=ssl_X509_STORE_create((char*)crlfile,(char*)crldir);
    if (!ncrl) {
	_pam_log(LOG_WARNING,"Can't load CA revocation lists!");
    }
//    x509st_CRL=ncrl;
    
    *ca=nca;
    *crl=ncrl;
    return 0;
}

static void SSLfree(X509_STORE *ca, X509_STORE *crl) {
    if (crl) X509_STORE_free(crl);
    if (ca) X509_STORE_free(ca);
}

static STACK_OF(X509) *X509sload(const char *cert_file) {
    STACK_OF(X509) *chain;
    X509 *cert;
    BIO* bio;
    
    
    chain = sk_X509_new_null();
    if (!chain) {
	_pam_log(LOG_ERR, "Error creating stack of certificates!");
	return NULL;
    }
    
    bio = BIO_new_file(cert_file, "rb");
    if (!bio) {
	_pam_log(LOG_ERR, "Error opening certificate file!");
	sk_X509_free(chain);
	return NULL;
    }
    
    while (1) {
#if FILE_FORMAT == SSL_FILETYPE_PEM
    	cert = PEM_read_bio_X509_AUX(bio, NULL, NULL, NULL);
#else 
	cert = d2i_X509_bio(bio, NULL);
#endif
	if (cert) {
	    if (sk_X509_push(chain,cert)<1) {
		_pam_log(LOG_WARNING, "Error pushing certificate to stack!");
	    }
	} else break;
    }
    return chain;
}

static void X509sfree(STACK_OF(X509) *chain) {
    if (chain) sk_X509_free(chain);
}

static X509 *X509load(const char *cert_file) {
    X509 *cert;
    BIO* bio;
    
    
    cert = X509_new();
    if (!cert) {
	_pam_log(LOG_ERR, "Error creating X509 certificate structure!");
	return NULL;
    }
    
    bio = BIO_new_file(cert_file, "rb");
    if (!bio) {
	_pam_log(LOG_ERR, "Error opening certificate file!");
	X509_free(cert);
	return NULL;
    }
    
#if FILE_FORMAT == SSL_FILETYPE_PEM
    cert = PEM_read_bio_X509_AUX(bio, NULL, NULL, NULL);
#else 
    cert = d2i_X509_bio(bio, NULL);
#endif
    if (!cert) {
	_pam_log(LOG_ERR, "Error reading certificate from file!");
    }
    return cert;
}

static void X509free(X509 *cert) {
    if (cert) X509_free(cert);
}

/*
static int get_key_password(char *buf, int size, int rwflag, void *userdata) {
    printf("Input PEM pass phrase: ");
    strncpy(buf,"argh13",size);buf[size-1]=0;
    return strlen(buf);
    return 0;
}
*/

char *authenticate(const char *user, pam_handle_t *pamh, const char *cadir, const char *cafile, const char *crldir, const char *crlfile, const char *keydir) {
    X509_STORE *ca;
    X509_STORE *crls;

    DIR *dir;
    struct dirent *de;
    struct stat st;
    char cert_file[MAX_FILENAME_SIZE+1];
    char key_file[MAX_FILENAME_SIZE+1];
    unsigned int dlen,len;

    SSL_CTX *ctx;

    STACK_OF(X509) *chain;
    X509 *cert;

    X509_STORE_CTX *sctx;
    
    int err;
    char *ruser;

    struct _pam_userdata pud;

    pud.pamh=pamh;
    
    if (SSLinit(&ca,&crls,cadir,cafile,crldir,crlfile)) return NULL;
    
    dlen=strlen(keydir)+1;
    if (dlen+8>MAX_FILENAME_SIZE) {
	_pam_log(LOG_ERR,"Key directory name to long!");
	SSLfree(ca,crls);
	return NULL;
    }
    sprintf(cert_file,"%s/",keydir);
    sprintf(key_file,"%s/",keydir);
    
    dir = opendir(keydir);
    if (!dir) {
	_pam_log(LOG_NOTICE,"Can't open key directory!");
	SSLfree(ca,crls);
	return NULL;
    }
    
    while ((de=readdir(dir))!=NULL) {
	len = strlen(de->d_name);

	if (strcmp(de->d_name+len-4,".crt")==0) {
	    if (len+dlen>MAX_FILENAME_SIZE) {
		_pam_log(LOG_NOTICE,"Key(%s) file name to long!",de->d_name);
		continue;
	    }
	    strcpy(cert_file+dlen,de->d_name);
	    strncpy(key_file+dlen,de->d_name,len-4);strcpy(key_file+dlen+len-4,".key");
		
		// We don't support links (It's _very_ insecure)!!! Check for regular files...	
	    if (lstat(cert_file,&st)) {
		_pam_log(LOG_NOTICE,"Certificate(%s) cann't be stated!",de->d_name);
		continue;
	    }
	    if (!S_ISREG(st.st_mode)) {
		_pam_log(LOG_NOTICE,"Certificate(%s) isn't regular file!",de->d_name);
		continue;
	    }
	    if (lstat(key_file,&st)) {
		_pam_log(LOG_NOTICE,"Key(%s) cann't be stated!",de->d_name);
		continue;
	    }
	    if (!S_ISREG(st.st_mode)) {
		_pam_log(LOG_NOTICE,"Key(%s) isn't regular file!",de->d_name);
		continue;
	    }
	    
		// Check if key and certificate correspond each other
	    ctx = SSL_CTX_new(SSLv23_method());
	    if (!ctx) {
		_pam_log(LOG_ERR, "Error initialising OpenSSL Context!");
		break;
	    }
	    
	    strncpy(pud.certname,de->d_name,len-4);pud.certname[len-4]=0;
	    SSL_CTX_set_default_passwd_cb(ctx,get_key_password);
	    SSL_CTX_set_default_passwd_cb_userdata(ctx,&pud);
	    
	    if (SSL_CTX_use_certificate_file(ctx, cert_file, FILE_FORMAT) != 1)  {
		_pam_log(LOG_NOTICE, "Error loading certificate(%s)!",de->d_name);
		SSL_CTX_free(ctx);
		continue;
	    }

	    if (SSL_CTX_use_PrivateKey_file(ctx, key_file, FILE_FORMAT) != 1)  {
		_pam_log(LOG_NOTICE, "Error loading private key(%s)!",de->d_name);
		SSL_CTX_free(ctx);
		continue;
	    }
	    
	    if (SSL_CTX_check_private_key(ctx)!=1) {
		_pam_log(LOG_NOTICE, "Private key doesn't correspond to certificate(%s)!",de->d_name);
		SSL_CTX_free(ctx);
		continue;
	    }
	    
	    SSL_CTX_free(ctx);

		// Check certificate
	    cert=X509load(cert_file);
	    if (!cert) continue;
	    
	    sctx=X509_STORE_CTX_new();
	    if (!sctx) {
		_pam_log(LOG_ERR, "OpenSSL STORE_CTX structure creation failed!");
		X509_free(cert);
		continue;
	    }
	    
	    if (X509_STORE_CTX_init(sctx,ca,cert,NULL)!=1) {
		_pam_log(LOG_ERR, "OpenSSL STORE_CTX structure initialization failed!");
		X509_STORE_CTX_free(sctx);
		X509_free(cert);
		continue;
	    }
	    if (X509_verify_cert(sctx)!=1) {
		_pam_log(LOG_NOTICE,"Certificate(%s) verification failed!", de->d_name);
		X509_STORE_CTX_free(sctx);
		X509_free(cert);
		continue;
	    }
	    if (verify_cb_CRL(crls,sctx)!=1) {
		_pam_log(LOG_NOTICE,"Certificate(%s) is revoked!", de->d_name);
		X509_STORE_CTX_free(sctx);
		X509_free(cert);
		continue;
	    }
	    X509_STORE_CTX_free(sctx);
// I have no idea howto best way organize chain authentication: Use first certificate in file,
// or may be use closest to CA certificate. Second can permit somebody to provide ability to
// third persons use his account without giving them his private key and password.
//	    chain=X509sload(cert_file);
//	    if (!chain) continue;
//	    cert=sk_X509_value(chain,0);
//	    X509sfree(chain);
// Find closest certificate, look: xmlSecOpenSSLX509StoreVerify(x509vfy.c, XMLSec 1.2)
// X509_STORE_CTX_init(&sctx,ca,auth_cert,chain)

		// Authenticate user and determine user name
	    if ((!user)||(!user[0])||(!strcmp(user,"x509detect"))) {
		ruser=(char*)malloc(MAX_USERNAME_SIZE);ruser[0]=0;
		if (!ruser) {
		    _pam_log(LOG_ERR,"Failed to allocate memory!");
		    X509_free(cert);
		    break;
		}
		err=ssl_x509_auth(cert,ruser);
	    } else {
		ruser=NULL;
		err=ssl_x509_auth(cert,user);
	    }
	    
	    if (err==1) {
		if (!ruser) {
		    ruser=(char*)malloc(strlen(user)+1);
		    if (!ruser) {
			_pam_log(LOG_ERR,"Failed to allocate memory!");
			X509_free(cert);
			break;
		    }
		    strcpy(ruser,user);
		}
		return ruser;
	    } else if (err<0) {
		_pam_log(LOG_ERR,"X509 authentication failed with error code %d!",err);
		if (ruser) free(ruser);
		X509free(cert);
		break;
	    } 

	    if (ruser) free(ruser);
	    X509free(cert);
	}
    }
    
    closedir(dir);
    SSLfree(ca,crls);
    return NULL;
}

/*
main() {
    char *res;
    res=authenticate("csa5",ca_dir,ca_file,crl_dir,crl_file,key_dir);
    if (res) puts(res);
    else puts("Failed");
}
*/
